﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using System.Runtime.InteropServices;
using System.IO.Ports;
using System.Threading;

namespace Controller {
    public partial class frmController : Form {
        #region Declarations
        ControlCode_t ControlCode = new ControlCode_t();

        private List<Keys> PressedKeys = new List<Keys>();
        private List<Keys> ConfiguredKeys = new List<Keys>() { Keys.Up, Keys.Down, Keys.Left, Keys.Right, Keys.Q, Keys.W, Keys.E, Keys.P };

        private Thread recursivelyThread = null;
        private ManualResetEvent recursivelyThreadStopEvent = null;
        private object recursivelySync = new object();

        static volatile byte[] recursivelyBuffer;
        static volatile int recursivelyInterval;
        static volatile bool LastRecursivelySend = false;
        static DateTime LastSentTime = DateTime.Now;

        private SerialPort serial = null;
        private string comport;
        private string baudrate;
        private StringBuilder serialBuffer = new StringBuilder();
        string strPostfix = "\n";
        #endregion

        #region Contructor
        public frmController() {
            InitializeComponent();
        }
        #endregion

        #region Events
        private void frmController_Load(object sender, EventArgs e) {
            int[] baudrates = {
                                    1200
                                    ,2400
                                    ,4800
                                    ,9600
                                    ,19200
                                    ,38400
                                    ,57600
                                    ,115200
                                    ,230400
                                    ,460800
                                    ,921600
                                    ,1382400
                                  };

            foreach (int i in baudrates) {
                cboBaudRate.Items.Add(Convert.ToString(i));
            }

            cboBaudRate.SelectedIndex = 6;

            ControlCode = new ControlCode_t() { headerType = 0x01 };
        }

        private void frmController_FormClosing(object sender, FormClosingEventArgs e) {
            Disconnect();
            btnConnect.Enabled = true;
            btnDisconnect.Enabled = false;
        }

        private void chkRemoteControl_CheckedChanged(object sender, EventArgs e) {
            if (chkRemoteControl.Checked) {
                foreach (Control control in this.Controls) {
                    control.PreviewKeyDown += new PreviewKeyDownEventHandler(Controls_PreviewKeyDown);
                }

                this.KeyPreview = true;
            }
            else {
                foreach (Control control in this.Controls) {
                    control.PreviewKeyDown -= Controls_PreviewKeyDown;
                }
                this.KeyPreview = false;
            }
        }

        private void Controls_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e) {
            e.IsInputKey = true;
        }

        private void frmController_KeyDown(object sender, KeyEventArgs e) {
            if (!ConfiguredKeys.Contains(e.KeyCode))
                return;

            if (!PressedKeys.Contains(e.KeyCode)) {
                PressedKeys.Add(e.KeyCode);

                SendCommand();
            }

            if (chkRemoteControl.Checked)
                e.SuppressKeyPress = true;
        }

        private void frmController_KeyUp(object sender, KeyEventArgs e) {
            if (!ConfiguredKeys.Contains(e.KeyCode))
                return;

            if (PressedKeys.Contains(e.KeyCode)) {
                PressedKeys.Remove(e.KeyCode);

                SendCommand();
            }

            if (chkRemoteControl.Checked)
                e.SuppressKeyPress = true;
        }

        private void cboComPort_DropDown(object sender, EventArgs e) {
            GetPorts();
        }

        private void cboDelimiter_SelectedIndexChanged(object sender, EventArgs e) {
            switch (cboDelimiter.Text) {
                case "\\n":
                    strPostfix = "\n";
                    break;
                case "\\r\\n":
                    strPostfix = "\r\n";
                    break;
                default:
                    strPostfix = "";
                    break;
            }
        }

        private void btnConnect_Click(object sender, EventArgs e) {
            try {
                if (cboComPort.Text == "") {
                    MessageBox.Show("Select the Port!");
                    return;
                }
                if (cboBaudRate.Text == "") {
                    MessageBox.Show("Select the Baud Rate!");
                    return;
                }

                comport = cboComPort.Text;
                baudrate = cboBaudRate.Text;

                if (Connect()) {
                    btnConnect.Enabled = false;
                    btnDisconnect.Enabled = true;
                }
            }
            catch (Exception ex) { txtMessagesAppend(ex.Message); }
        }

        private void btnDisconnect_Click(object sender, EventArgs e) {
            try {
                Disconnect();
                btnConnect.Enabled = true;
                btnDisconnect.Enabled = false;
            }
            catch (Exception ex) { txtMessagesAppend(ex.Message); }
        }

        private void serial_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e) {
            try {
                while (serial.BytesToRead > 0) {
                    char newchar = (char)serial.ReadByte();
                    serialBuffer.Append(newchar);

                    if (strPostfix != "" && serialBuffer.ToString().Contains(strPostfix)) {
                        if (serialBuffer.ToString().Contains("PONG")) {
                            System.TimeSpan diff = DateTime.Now - LastSentTime;
                            txtMessagesAppend(">" + serialBuffer.ToString().Replace(strPostfix, "") + " it took " + diff.TotalMilliseconds.ToString());
                        }
                        else {
                            txtMessagesAppend(">" + serialBuffer.ToString().Replace("\n", "").Replace("\r", ""));
                        }

                        serialBuffer.Clear();
                    }
                }

                if (strPostfix == "" && serialBuffer.Length > 0) {
                    txtMessagesAppend(">" + serialBuffer.ToString().Replace("\n", "").Replace("\r", ""));
                    serialBuffer.Clear();
                }
            }
            catch (Exception ex) { txtMessagesAppend("Data Received: " + ex.Message + ", Stack " + ex.StackTrace.Trim()); }
        }
        #endregion

        #region Methods and Functions
        private void txtMessagesAppend(string strMessage) {
            string Message = DateTime.Now.ToString("dd/MM/HH:mm:ss.ffff") + ": " + strMessage + "\r\n";

            if (txtMessages.InvokeRequired) {
                txtMessages.Invoke((MethodInvoker)delegate {
                    txtMessages.AppendText(Message);
                });
            }
            else {
                txtMessages.AppendText(Message);
            }
        }

        private void GetPorts() {
            try {
                cboComPort.Items.Clear();

                SerialPort serialport = null;
                foreach (string str in SerialPort.GetPortNames()) {
                    serialport = new SerialPort(str);
                    if (serialport.IsOpen == false)
                        cboComPort.Items.Add(str);
                }

            }
            catch (Exception ex) { txtMessagesAppend("GetPorts: " + ex.Message); }
        }

        private bool Connect() {
            try {
                if (serial == null) {
                    serial = new SerialPort();
                    serial.PortName = comport;
                    serial.BaudRate = Convert.ToInt32(baudrate);

                    serial.Parity = Parity.None;
                    serial.StopBits = StopBits.One;
                    serial.DataBits = 8;
                    //serial.Handshake = Handshake.XOnXOff;

                    serial.ReadTimeout = 2000;
                    serial.WriteTimeout = 2000;
                    serial.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(serial_DataReceived);
                }

                if (serial != null) {
                    if (serial.IsOpen == false)
                        serial.Open();

                    if (serial.IsOpen) {
                        this.Text = "Controller Serial: " + comport + ", " + baudrate + " - " + "Connected";
                        txtMessagesAppend("Connected");
                        return true;
                    }
                }
                return false;
            }
            catch (SystemException ex) {
                txtMessagesAppend(ex.Message);

                if (serial != null) {
                    if (serial.IsOpen)
                        serial.Close();

                    serial.Dispose();
                    serial = null;
                }
                return false;
            }
        }

        private void Disconnect() {
            try {
                if (serial != null) {
                    recursivelyStop();
                    serial.DataReceived -= serial_DataReceived;

                    if (serial.IsOpen) {
                        serial.Close();
                        this.Text = comport + " " + baudrate + " - " + "Disconnected";

                        txtMessagesAppend("Disconnected");
                    }
                }
                serial = null;
            }
            catch (Exception ex) { txtMessagesAppend(ex.Message); }
        }

        private void recursivelyThreadWorker() {
            try {
                DateTime SentTime = DateTime.Now;
                while (!recursivelyThreadStopEvent.WaitOne(10)) {
                    if (recursivelyInterval > 0) {
                        System.TimeSpan diff = DateTime.Now - LastSentTime;

                        if (diff.TotalMilliseconds > recursivelyInterval) {
                            txtMessagesAppend("Bytes recursively");
                            Send(recursivelyBuffer);
                            if (LastRecursivelySend)
                                recursivelyInterval = 0;
                        }
                    }
                }
            }
            catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
            finally {
            }
        }

        private void recursivelyStop() {
            lock (recursivelySync) {
                if (recursivelyThread != null) {
                    recursivelyInterval = 0;
                    recursivelyThreadStopEvent.Set();

                    while (recursivelyThread.IsAlive) {
                    }

                    recursivelyThread = null;

                    if (recursivelyThreadStopEvent != null) {
                        recursivelyThreadStopEvent.Close();
                    }
                }
            }
        }

        private void Send(string strCommand, string Postfix = "") {
            try {
                if (serial == null || !serial.IsOpen) {
                    txtMessagesAppend("Connection is Closed!");
                    return;
                }

                char[] spSend = (strCommand + (Postfix == "" ? strPostfix : Postfix)).ToCharArray();
                serial.Write(spSend, 0, spSend.Length);

                lock (recursivelySync) {
                    LastSentTime = DateTime.Now;
                }

                txtMessagesAppend(strCommand);
            }
            catch (Exception ex) { txtMessagesAppend("Send: " + ex.Message); }
        }

        private void Send(byte[] Bytes) {
            try {
                if (serial == null || !serial.IsOpen) {
                    txtMessagesAppend("Connection is Closed!");
                    return;
                }

                serial.Write(Bytes, 0, Bytes.Length);

                lock (recursivelySync) {
                    LastSentTime = DateTime.Now;
                }

                StringBuilder sb = new StringBuilder();

                for (int i = 0; i < Bytes.Length; i++) {
                    sb.Append(" byte" + i.ToString() + ": " + printBits(Bytes[i]));
                }

                txtMessagesAppend(sb.ToString());
            }
            catch (Exception ex) { txtMessagesAppend("Send: " + ex.Message); }
        }

        private void Send(byte[] Buffer, int interval) {
            if (interval == 0) {
                lock (recursivelySync) {
                    //Send if recursivelyThread is not sending
                    if (recursivelyInterval == 0) {
                        Send(Buffer);
                    }
                    else {
                        //set recursivelyBuffer 
                        recursivelyBuffer = Buffer;
                        //set to send last time
                        LastRecursivelySend = true;
                    }
                }
            }
            else {
                lock (recursivelySync) {
                    LastRecursivelySend = false;
                    recursivelyBuffer = Buffer;
                    recursivelyInterval = interval;

                    if (recursivelyThread == null || !recursivelyThread.IsAlive) {
                        LastSentTime = DateTime.Now.AddMilliseconds(-interval);
                        recursivelyThreadStopEvent = new ManualResetEvent(false);
                        recursivelyThread = new Thread(new ThreadStart(recursivelyThreadWorker));
                        recursivelyThread.Name = "recursivelyThread";
                        recursivelyThread.Start();
                    }
                }
            }
        }

        private void SendCommand() {
            bool recursively = false;
            ControlCode.Clear();
            ControlCode.headerType = 0x01;
            ControlCode.Type = 0x01;

            if (PressedKeys.Count > 0) {
                foreach (Keys key in PressedKeys.ToArray()) {
                    txtMessagesAppend(key.ToString());
                }
            }
            else
                txtMessagesAppend("No Keys are pressed");

            if (PressedKeys.Contains(Keys.P)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 0);//P1-1 POWER
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 0);

            if (PressedKeys.Contains(Keys.W)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 1);//P1-2 FIRE
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 1);

            if (PressedKeys.Contains(Keys.E)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 2);//P1-3 TURRET RIGHT
                recursively = true;
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 2);

            if (PressedKeys.Contains(Keys.Q)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 3);//P1-3 TURRET LEFT
                recursively = true;
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 3);

            //D-PAD
            if (PressedKeys.Contains(Keys.Up)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 4);//P1-4 UP
                recursively = true;
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 4);

            if (PressedKeys.Contains(Keys.Down)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 5); //P1-5 DOWN
                recursively = true;
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 5);

            if (PressedKeys.Contains(Keys.Left)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 6); //P1-6 LEFT
                recursively = true;
            }
            else
                ControlCode.b1 = clearBit(ControlCode.b1, 6);

            if (PressedKeys.Contains(Keys.Right)) {
                ControlCode.b1 = SetBit(ControlCode.b1, 7); //P1-7 RIGHT
                recursively = true;
            }
            else {
                ControlCode.b1 = clearBit(ControlCode.b1, 7);
            }

            txtMessagesAppend(printBits(ControlCode.b1));

            if (recursively) {
                SendWithCRC(false, 500);//should be better send up from 200 mileseconds becouse of wifi latency
            }
            else {
                SendWithCRC(false, 0);
            }
        }

        private void SendWithCRC(bool sendStop, int interval) {
            //TODO: send to net sockect
            byte[] data = ControlCode.Serialize();

            ushort crc = getCRCMODBUS(data, (ushort)(data.Length - 2));
            data[data.Length - 1] = (byte)((crc & 0xFF00) >> 8);
            data[data.Length - 2] = (byte)(crc & 0x00FF);

            if (serial != null)
                Send(data, interval);

            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < data.Length; i++) {
                sb.Append(" byte" + i.ToString() + ": " + printBits(data[i]));
            }
            txtMessagesAppend("SendWithCRC: " + sb.ToString() + " CRC: " + crc.ToString());

            if (sendStop) {
                byte lastCtrl = ControlCode.Type;
                ControlCode.Clear();
                ControlCode.Type = lastCtrl;
                SendWithCRC(false, 0);//KeepAliveMillis);
            }
        }

        private ushort getCRCMODBUS(byte[] buf, ushort len) {
            ushort crc = 0xFFFF;
            //ushort val = 0;

            for (ushort pos = 0; pos < len; pos++) {
                crc ^= (ushort)(0x00ff & buf[pos]);  // FIX HERE -- XOR byte into least sig. byte of crc

                for (int i = 8; i != 0; i--) {    // Loop over each bit
                    if ((crc & 0x0001) != 0) {      // If the LSB is set
                        crc >>= 1;                    // Shift right and XOR 0xA001
                        crc ^= 0xA001;
                    }
                    else                            // Else LSB is not set
                        crc >>= 1;                    // Just shift right
                }
            }
            // Note, crc has low and high bytes swapped, so use it accordingly (or swap bytes)
            //val = (ushort)((crc & 0xff) << 8);
            //val = (ushort)(val + ((crc >> 8) & 0xff));
            //System.out.printf("Calculated a CRC of 0x%x, swapped: 0x%x\n", crc, val);
            //return val;
            return crc;
        }

        private byte SetBit(byte self, int index) {
            byte mask = (byte)(1 << index);
            self = (byte)(self | mask);
            return self;
        }

        private byte clearBit(byte self, int index) {
            byte mask = (byte)(1 << index);
            self = (byte)(self & ~mask);
            return self;
        }

        private string printBits(byte myByte) {
            string sReturn = "";
            byte mask;
            for (mask = 0x80; mask != 0; mask >>= 1) {
                if ((mask & myByte) != 0)
                    sReturn += "1";
                else
                    sReturn += "0";
            }
            return sReturn;
        }
        #endregion

        #region Struct
        [Serializable()]
        [StructLayout(LayoutKind.Sequential, Size = 10, Pack = 1)]
        public struct ControlCode_t {
            public byte headerType;
            public byte Type;
            public byte b1;
            public byte b2;
            public byte b3;
            public byte b4;
            public byte b5;
            public byte b6;
            public ushort CRC;

            public byte[] Serialize() {
                // Get the byte size of a structure if it is to be
                // marshaled to unmanaged memory.
                Int32 iSizeOMyDataStruct = Marshal.SizeOf(this /*typeof(MyDataStruct)*/);
                // Allocate a byte array to contain the bytes of the unmanaged version
                // of the structure.
                byte[] byteArrayMyDataStruct = new byte[iSizeOMyDataStruct];
                // Allocate a GCHandle to pin the array
                // in memory in order to obtain its pointer.
                GCHandle gch = GCHandle.Alloc(byteArrayMyDataStruct, GCHandleType.Pinned);
                try {
                    // Obtain a pointer to the byteArrayMyDataStruct array in memory.
                    IntPtr pbyteArrayMyDataStruct = gch.AddrOfPinnedObject();
                    // Copy all bytes from the managed MyDataStruct structure into
                    // the byte array.
                    Marshal.StructureToPtr(this, pbyteArrayMyDataStruct, false);
                    // Return the byte array.
                    // It contains the serialized bytes of the MyDataStruct structure.
                    return byteArrayMyDataStruct;
                }
                catch (Exception e) {
                    MessageBox.Show(e.Message);
                    return null;
                }
                finally {
                    // Unpin the byteArrayMyDataStruct array in memory.
                    gch.Free();
                }
            }

            public void Deserialize(ref byte[] byteSerializedData) {
                // Allocate a GCHandle of the Pinned Type
                // for the byteSerializedData byte array.
                // This is possible for a byte array
                // because it is a blittable type.
                GCHandle gch = GCHandle.Alloc(byteSerializedData, GCHandleType.Pinned);
                try {
                    // Get a pointer to the byteSerializedData array.
                    IntPtr pbyteSerializedData = gch.AddrOfPinnedObject();
                    // Convert the array data of byteSerializedData
                    // directly into a MyDataStruct structure.
                    // The interop marshaler will use the MarshalAsAttribute
                    // of the fields of the MyDataStruct structure to
                    // perform field data conversions.
                    Marshal.PtrToStructure(pbyteSerializedData, this/*typeof(MyDataStruct)*/);
                }
                catch (Exception e) {
                    MessageBox.Show(e.Message);
                }
                finally {
                    // Free the GCHandle.
                    gch.Free();
                }
            }

            public void Clear() {
                this.headerType = 0x00;
                this.Type = 0x00;
                this.b1 = 0x00;
                this.b2 = 0x00;
                this.b3 = 0x00;
                this.b4 = 0x00;
                this.b5 = 0x00;
                this.b6 = 0x00;
                this.CRC = 0x0000;
            }
        }
        #endregion
    }
}
